// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file Boundary.h */

#ifndef AMDIS_BOUNDARY_H
#define AMDIS_BOUNDARY_H

#include "Global.h"
#include "MemoryManager.h"
#include <map>

namespace AMDiS {

  template<typename T> class WorldVector;

  /** \brief
   * Contains boundary constants
   */
  typedef enum {
    INTERIOR=0,   /**< interior => no boundary (b = 0) */
    DIRICHLET=1,  /**< dirichlet boundary (b > 0) */
    NEUMANN=-1,   /**< neumann boundary  (-100 < b < 0) */
    ROBIN=-100    /**< robin boundary ( b <= -100) */
  } BoundaryConstants;

  /** \brief
   * Type specifier for the different boundary types 
   */
  typedef signed int BoundaryType;

  BoundaryType newBound(BoundaryType oldBound, BoundaryType newBound);

}

#endif  // !_Boundary_H_ 
