// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file RefinementManager3d.h */

#ifndef AMDIS_REFINEMENT_MANAGER_3D_H
#define AMDIS_REFINEMENT_MANAGER_3D_H

#include "MemoryManager.h"

namespace AMDiS {

  // ============================================================================
  // ===== class RefinementManager3d ============================================
  // ============================================================================

  /** \ingroup Adaption 
   * \brief
   * Implements a RefinementManager for 3-dimensional meshes.
   */
  class RefinementManager3d : public RefinementManager
  {
  public:
    MEMORY_MANAGED(RefinementManager3d);

    /** \brief
     * Calls base class constructor.
     */
    RefinementManager3d() 
      : RefinementManager()
    {};

    /** \brief
     * destructor 
     */
    virtual ~RefinementManager3d() {};

  protected:
    /** \brief
     * Used by \ref setNewCoords
     */
    int newCoordsFct(ElInfo *el_info);

    /** \brief
     * Implements RefinementManager::setNewCoords
     */
    void setNewCoords();

    /** \brief
     *  gets the elements around the refinement edge with vertices  node[0] and
     *  node[1] ; refines those elements at this edge are not compatible 
     *  devisible;  			                                 
     *  dir determines the direction of the first neighbour
     *  get_refine_patch returns  1  if the domain's boundary is reached while
     *  looping around the refinement edge, otherwise  0
     */
    int getRefinePatch(ElInfo **el_info, DegreeOfFreedom *edge[2], int dir,
		       RCNeighbourList* refineList, int *n_neigh);

    /** \brief
     * Refines all elements in the patch.
     */
    DegreeOfFreedom refinePatch(DegreeOfFreedom *edge[2], RCNeighbourList* refineList,
				int n_neigh, bool bound);

    /** \brief
     * Implements RefinementManager::refineFunction.
     */
    ElInfo* refineFunction(ElInfo* el_info);

    /** \brief
     * Refines one Tetrahedron.
     */
    void bisectTetrahedron(RCNeighbourList* refineList, int index,
			   DegreeOfFreedom *dof[3], DegreeOfFreedom *edge[2]);

    /** \brief
     * Used by \ref bisectTetrahedron
     */
    void fillPatchConnectivity(RCNeighbourList* ref_list, int index);

  private:
    /** \brief
     * Refinement patch
     */
    static RCNeighbourList* refList;
  };

}

#endif // AMDIS_REFINEMENT_MANAGER_3D_H
